require 'flog_cli'
require 'date'

class RamperFileMetrics

  def initialize(flags)
    if flags.empty?
      selected = Array.new(CSV_COLS.count, true)
    else
      selected    = Array.new(CSV_COLS.count, false)
      selected[0] = true
      #TODO: use command line parser
      #TODO: change array to hash
      flags.each do |f|
        case f
          when '-a'
            selected[3] = true
          when '-c'
            selected[2] = true
          when '-f'
            selected[4] = true
          when '-l'
            selected[1] = true
        end
      end
    end

    @csv_header       = CSV_COLS.select.with_index { |_, ind| selected[ind] }
    @selected_metrics = METRICS.select.with_index { |_, ind| selected[ind] }
  end

  def process_file_format
    @csv_header.flatten.join(', ')
  end

  def process_file(file)
    @selected_metrics.map.with_index do |m, ind|
      begin
        m.call(file)
      rescue => e
        Array.new(@csv_header[ind].count, "Error #{e.message}".gsub(",", ";"))
      end
    end.flatten.join(', ')
  end

  private

  def self.start_end_dates_commits(file)
    dates       = `git log -- #{file} | grep 'Date: '`.split("\n")
    start_date  = DateTime.parse(dates.last.strip.gsub(/Date:   /, ''))
    end_date    = DateTime.parse(dates.first.strip.gsub(/Date:   /, ''))
    num_commits = dates.count

    [start_date, end_date, num_commits]
  end

  def self.lines_of_code(file)
    `wc -l #{file}`.split(" ").first.strip
  end

  def self.get_git_author_stats(file)
    `git log --pretty=short --no-merges -- #{file} | git shortlog -sne`.split("\n")
  end

  def self.author_stats(file)
    authors            = get_git_author_stats(file)
    author_table       = authors.map { |line| /\s+(\d+)\t(.*) <(.*)>/.match(line) { |_| { :count => $1.to_i, :name => $2, :email => $3 } } }
    unique_email_table = combine_authors(author_table, :email)
    unique_table       = combine_authors(unique_email_table, :name)
    num_authors        = unique_table.count
    top_author         = unique_table.max { |a, b| a[:count] <=> b[:count] }
    top_author         = "#{top_author[:name]} <#{top_author[:email]}>"

    [top_author, num_authors]
  end

  def self.combine_authors(original_table, combine_key)
    other_key  = combine_key == :name ? :email : :name
    result_map = Hash.new { |h, k| h[k]=[] }
    original_table.each { |row| result_map[row[combine_key]].push(row.reject { |k, _| k==combine_key }) }
    result_map.map { |key, val| { :count      => val.inject(0) { |tot_count, x| tot_count + x[:count] },
                                  combine_key => key,
                                  other_key   => val.max { |a, b| a[:count] <=> b[:count] }[other_key] } }
  end

  def self.flog_file(file)
    if /.*\.rb/.match(file)
      flogger = FlogCLI.new
      flogger.flog(file)
      flog_tot = flogger.total_score.round(2)
      flog_avg = flogger.average.round(2)

      [flog_tot, flog_avg]
    else
      Array.new(2, "N/A")
    end
  end

  # self.method blows up unless it is after the functions are defined
  CSV_COLS = ['File', 'LOC', ['First edit', 'Most recent edit', '# of commits'],
              ['Top author', '# of authors'], ['Flog score', 'Flog/method average']]
  METRICS  = [Proc.new { |x| x }, self.method(:lines_of_code), self.method(:start_end_dates_commits),
              self.method(:author_stats), self.method(:flog_file)]
end